import {
  DEFAULT_TIME_STEP,
  MINIMUM_VELOCITY,
  VISCOSITY_FACTOR,
  TIME_DILATION_FACTOR,
  MAXIMUM_STEPS,
  TRANSLATION_VELOCITY_FACTOR,
  ARROWHEAD_LENGTH,
  ARROWHEAD_WIDTH,
  MINIMUM_STEPS,
  CENTRAL_RESTORING_FORCE,
} from '../../util/constants.js';
import { Edge } from '../supporting/edge.js';
import { WeightedDirectedGraph } from '../supporting/wdg.js';
import { Box } from './box.js';
import { Vector, Rectangle } from '../supporting/geometry/index.js';
import { overlapRepulsionForce, targetRadiusForce } from './pairwise-forces.js';
import { Vertex } from '../supporting/vertex.js';

// Render children with absolute css positioning.

// Let there be a force between elements such that the force between
// any two elements is along the line between their centers,
// so that the elements repel when too close but attract when too far.

// The equilibrium distance can be tuned, e.g. can be scaled by an input.

// NOTE: (with optional overlay preferring a grid or some other shape?),

// NOTE: Could also allow user input dragging elements.
// What might be neat here is to implement a force-based resistance effect;
// basically, the mouse pointer drags the element with a spring rather than directly.
// If the shape of the graph resists the transformation,
// the distance between the element and the cursor should increase.

// On an interval, compute forces among the elements.
// Simulate the effects of these forces

// NOTE: Impart random nudges, and resolve their effects to a user-visible resolution
// before rendering.

// NOTE: When mouse is in our box, we could hijack the scroll actions to zoom in/out.

export class ForceDirectedGraph extends WeightedDirectedGraph {
  constructor(name, parentEl, { width = 800, height = 600, ...options } = {}) {
    super(name, options);
    this.box = new Box(name, parentEl, options);
    this.box.addClass('fixed');
    this.box.addClass('force-directed-graph');
    this.intervalTask = null;
    this.canvas = window.document.createElement('canvas');
    this.width = width;
    this.height = height;
    this.box.el.style.width = `${width}px`;
    this.box.el.style.height = `${height}px`;
    this.box.el.appendChild(this.canvas);
    this.fitCanvasToGraph();
    this.nodes = [];
    this.edges = [];
  }

  fitCanvasToGraph() {
    [this.canvas.width, this.canvas.height] = this.box.rect.dimensions;
  }

  addVertex(...args) {
    const vertex = super.addVertex(...args);
    const box = this.box.addBox(vertex.id);

    // Link from the graph vertex to the corresponding display box
    vertex.box = box;

    // Link from the display box to the corresponding graph vertex
    box.vertex = vertex;

    box.addClass('absolute');
    box.addClass('vertex');
    box.move(this.box.rect.dimensions.scale(0.5));
    box.velocity = Vector.from([0, 0]);
    box.setInnerHTML(vertex.getHTML());
    this.nodes.push(box);

    // When vertex properties are updated, re-render the node contents
    vertex.onUpdate = () => {
      box.setInnerHTML(vertex.getHTML());
      // Maybe resolve forces
      this.runUntilEquilibrium();
    };

    this.runUntilEquilibrium();

    // Allow moving vertices with the mouse
    box.el.addEventListener('mousedown', (e) => {
      console.log('mousedown, button:', e.button);
      if (!this.mouseMoving) {
        e.preventDefault();
        // Record current mouse position
        this.mousePosition = Vector.from([e.clientX, e.clientY]);
        // Begin tracking mouse movement
        this.mouseMoving = box;
      }
    });
    document.addEventListener('mousemove', (e) => {
      if (this.mouseMoving === box) {
        const mousePosition = Vector.from([e.clientX, e.clientY]);
        // Apply translation
        box.move(mousePosition.subtract(this.mousePosition));
        // Update current mouse position
        this.mousePosition = mousePosition;
        // Equilibrate
        this.runUntilEquilibrium();
      }
    });
    document.addEventListener('mouseup', () => {
      // Stop tracking mouse movement
      this.mouseMoving = null;
      // Equilibrate
      this.runUntilEquilibrium();
    });

    return vertex;
  }

  addEdge(type, from, to, ...rest) {
    from = from instanceof Vertex ? from : this.getVertex(from);
    to = to instanceof Vertex ? to : this.getVertex(to);
    if (!from) throw new Error(`from ${from}: Node not found`);
    if (!to) throw new Error(`to ${to}: Node not found`);

    const edge = super.addEdge(type, from, to, ...rest);
    const box = this.box.addBox(Edge.getKey({ from, to, type }));

    edge.box = box;
    box.edge = edge;

    box.addClass('absolute');
    box.addClass('edge');
    box.setInnerHTML(edge.getHTML());
    this.edges.push(box);

    // Initially place near the midpoint between the `from` and `to` nodes.
    const midpoint = from.box.rect.center.add(to.box.rect.center).scale(0.5);
    const startPosition = midpoint.subtract(box.rect.dimensions.scale(0.5));
    box.move(startPosition);
    box.velocity = Vector.from([0, 0]);
    this.runUntilEquilibrium();

    return edge;
  }

  setEdgeWeight(...args) {
    const edge = super.setEdgeWeight(...args);
    edge.displayEdgeNode();
    return edge;
  }

  drawEdgeLines() {
    const ctx = this.canvas.getContext('2d');
    ctx.clearRect(0, 0, this.width, this.height);
    ctx.strokeStyle = '#57747d';
    ctx.fillStyle = '#57747d';
    for (const { edge } of this.edges) {
      ctx.beginPath();
      ctx.moveTo(...edge.from.box.rect.center);
      ctx.lineTo(...edge.box.rect.center);
      ctx.stroke();

      const direction = edge.to.box.rect.center.subtract(edge.box.rect.center);
      const arrowTail = edge.box.rect.lineIntersect(edge.to.box.rect.center, direction.scale(-1));
      const arrowPoint = edge.to.box.rect.lineIntersect(edge.box.rect.center, direction);
      const arrowBaseCenter = arrowPoint.subtract(direction.normalize().scale(ARROWHEAD_LENGTH));
      const arrowBaseDirection = Vector.from([direction[1], -direction[0]]).normalize();
      const arrowBaseLeft = arrowBaseCenter.add(arrowBaseDirection.scale(ARROWHEAD_WIDTH));
      const arrowBaseRight = arrowBaseCenter.subtract(arrowBaseDirection.scale(ARROWHEAD_WIDTH));

      ctx.beginPath();
      ctx.moveTo(...arrowTail);
      ctx.lineTo(...arrowPoint);
      ctx.stroke();

      ctx.beginPath();
      ctx.moveTo(...arrowPoint);
      ctx.lineTo(...arrowBaseLeft);
      ctx.lineTo(...arrowBaseRight);
      ctx.closePath();
      ctx.fill();
    }
  }

  async runUntilEquilibrium(tDelta = DEFAULT_TIME_STEP) {
    this.steps = 0;
    if (this.intervalTask) {
      return Promise.resolve();
    }
    return new Promise((resolve, reject) => {
      this.intervalTask = setInterval(() => {
        if (this.intervalTaskExecuting) return;
        this.steps++;
        if (this.steps > MAXIMUM_STEPS) {
          clearInterval(this.intervalTask);
          this.intervalTask = null;
          reject(new Error('Exceeded max steps to reach equilibrium'));
        }
        this.intervalTaskExecuting = true;
        const { atEquilibrium } = this.computeEulerFrame(tDelta);
        this.intervalTaskExecuting = false;
        if (atEquilibrium) {
          console.log(`Reached equilibrium after ${this.steps} steps`);
          clearInterval(this.intervalTask);
          this.intervalTask = null;
          resolve();
        }
      }, tDelta * TIME_DILATION_FACTOR);
    });
  }

  computeEulerFrame(tDelta = DEFAULT_TIME_STEP) {
    // Compute net forces on each box in the graph
    const boxes = [...this.nodes, ...this.edges];

    // Initialize net force vectors
    for (const box of boxes) {
      box.forces = [];
    }

    const addForce = (box, force, type) => box.forces.push({ force, type });

    // All boxes repel each other if they overlap
    for (const boxA of boxes) {
      const idxA = boxes.indexOf(boxA);
      for (const boxB of boxes.slice(idxA + 1)) {
        const force = overlapRepulsionForce(boxA, boxB);
        addForce(
          boxA,
          force.scale(-1),
          `${boxB.name} -- overlapRepulsion --> ${boxA.name}`,
        );
        addForce(
          boxB,
          force,
          `${boxA.name} -- overlapRepulsion --> ${boxB.name}`,
        );
      }
    }

    // Center of graph attracts all boxes that are outside the graph
    for (const box of boxes) {
      if (!this.box.rect.doesContain(box.rect.center)) {
        const r = this.box.rect.center.subtract(box.rect.center);
        addForce(
          box,
          r.normalize().scale(CENTRAL_RESTORING_FORCE),
          `center -- attraction --> ${box.name}`,
        );
      }
    }

    // Compute edge-related forces
    for (const edgeBox of this.edges) {
      const { edge } = edgeBox;
      const fromBox = edge.from.box;
      const toBox = edge.to.box;

      // Attraction to the `from` and `to` nodes
      addForce(
        edgeBox,
        targetRadiusForce(fromBox, edgeBox, 0),
        `${edgeBox.name} -- attract fromBox --> ${fromBox.name}`,
      );
      addForce(
        edgeBox,
        targetRadiusForce(toBox, edgeBox, 0),
        `${edgeBox.name} -- attract toBox --> ${toBox.name}`,
      );

      // Pairwise force between nodes
      {
        const force = targetRadiusForce(fromBox, toBox);
        addForce(
          fromBox,
          force.scale(-1),
          `${toBox.name} -- pairwise targetRadius --> ${fromBox.name}`,
        );
        addForce(
          toBox,
          force,
          `${fromBox.name} -- pairwise targetRadius --> ${toBox.name}`,
        );
      }
    }

    // Do not apply forces to a box if it is being moved by the mouse
    for (const box of boxes) {
      if (this.mouseMoving === box) {
        box.netForce = Vector.zeros(2);
      } else {
        box.netForce = box.forces.reduce((net, { force }) => net.add(force), Vector.from([0, 0]));
      }
    }

    // Compute velocities
    for (const box of boxes) {
      box.velocity = box.velocity.add(box.netForce.scale(tDelta));
      // Apply some drag
      box.velocity = box.velocity.scale(1 - VISCOSITY_FACTOR);
    }

    // When all velocities are below MINIMUM_VELOCITY and we have executed more than
    // MINIMUM_STEPS, we have reached equilibrium.
    let atEquilibrium = this.steps > MINIMUM_STEPS;

    // Apply velocities
    for (const box of boxes) {
      if (box.velocity.magnitude >= MINIMUM_VELOCITY) {
        atEquilibrium = false;
        box.move(box.velocity);
      }
    }

    // Center the items by computing the bounding box and centering that
    if (!this.mouseMoving) {
      const topLeft = Vector.from(boxes[0].position);
      const bottomRight = Vector.from(boxes[0].position);
      for (const box of boxes) {
        for (const vertex of box.rect.vertices) {
          topLeft[0] = Math.min(topLeft[0], vertex[0]);
          topLeft[1] = Math.min(topLeft[1], vertex[1]);

          bottomRight[0] = Math.max(bottomRight[0], vertex[0]);
          bottomRight[1] = Math.max(bottomRight[1], vertex[1]);
        }
      }
      const boundingRect = new Rectangle(topLeft, bottomRight.subtract(topLeft));
      const graphCenter = Vector.from([this.width, this.height]).scale(0.5);
      const offset = graphCenter.subtract(boundingRect.center);
      const translate = offset.scale(TRANSLATION_VELOCITY_FACTOR);

      if (translate.magnitude >= MINIMUM_VELOCITY) {
        atEquilibrium = false;

        // Apply translations
        for (const box of boxes) {
          box.move(translate);
        }
      }
    }

    // TODO: Scaling to fit

    this.drawEdgeLines();

    return { atEquilibrium };
  }
}
