import { Box } from '../../classes/display/box.js';
import { ForceDirectedGraph } from '../../classes/display/force-directed.js';
import { Rectangle, Vector } from '../../classes/supporting/geometry/index.js';
import { overlapRepulsionForce, targetRadiusForce } from '../../classes/display/pairwise-forces.js';
import { delayOrWait } from '../../classes/display/scene-controls.js';
import { Scene } from '../../classes/display/scene.js';
import { DEFAULT_OVERLAP_BUFFER, EPSILON } from '../../util/constants.js';
import { mochaRun } from '../../util/helpers.js';

const rootElement = document.getElementById('scene');
const rootBox = new Box('rootBox', rootElement).flex();
window.scene = new Scene('WDG test', rootBox);

describe('Force-Directed Graph', function tests() {
  this.timeout(0);

  let graph;

  before(() => {
    graph = (window.graph = new ForceDirectedGraph('test1', window.scene.middleSection.el, {
      width: 1200, height: 900,
    }));

    graph.addVertex('v1', 'box1');
  });

  it('rectangle should be a polygon with 4 vertices', () => {
    const rect = new Rectangle([0, 0], [1, 1]);
    rect.vertices[0].should.eql([0, 0]);
    rect.vertices[1].should.eql([1, 0]);
    rect.vertices[2].should.eql([1, 1]);
    rect.vertices[3].should.eql([0, 1]);
  });

  it('should measure extent of overlap', () => {
    const rects = [
      new Rectangle([0, 0], [1, 1]),
      new Rectangle([0.5, 0], [1, 1]),
      new Rectangle([0, 0.5], [1, 1]),
      new Rectangle([0.5, 0.5], [1, 1]),
      new Rectangle([0, 1], [1, 1]),
      new Rectangle([0, 0], [1, 2]),
      new Rectangle([0, 0], [0.5, 0.5]),
      new Rectangle([2, 2], [1, 1]),
    ];
    rects[0].doesOverlap(rects[1]).should.eql(1);
    rects[0].doesOverlap(rects[2]).should.eql(1);
    rects[0].doesOverlap(rects[3]).should.eql(0.5);
    rects[0].doesOverlap(rects[4]).should.eql(0);
    rects[0].doesOverlap(rects[5]).should.eql(1);
    rects[0].doesOverlap(rects[6]).should.eql(1);
    rects[0].doesOverlap(rects[7]).should.eql(0);

    rects[1].doesOverlap(rects[0]).should.eql(1);
    rects[2].doesOverlap(rects[0]).should.eql(1);
    rects[3].doesOverlap(rects[0]).should.eql(0.5);
    rects[4].doesOverlap(rects[0]).should.eql(0);
    rects[5].doesOverlap(rects[0]).should.eql(1);
    rects[6].doesOverlap(rects[0]).should.eql(1);
    rects[7].doesOverlap(rects[0]).should.eql(0);
  });

  it('overlapping boxes should repel', () => {
    const rect1 = new Rectangle([0, 0], [1, 1]);
    const rect2 = new Rectangle([0, 0], [1, 2]);
    rect1.center.should.eql([0.5, 0.5]);
    rect2.center.should.eql([0.5, 1]);
    const force1 = overlapRepulsionForce(rect1, rect2, 10);
    force1.should.eql([0, 10]);
  });

  it('adjacent boxes should repel', () => {
    const rect1 = new Rectangle([0, 0], [1, 1]);
    const rect2 = new Rectangle([0, 1], [1, 1]);
    const rect3 = new Rectangle([0, DEFAULT_OVERLAP_BUFFER / 2 + 1], [1, 1]);
    const rect4 = new Rectangle([0, DEFAULT_OVERLAP_BUFFER + 1], [1, 1]);
    const rect5 = new Rectangle([DEFAULT_OVERLAP_BUFFER + 1, DEFAULT_OVERLAP_BUFFER + 1], [1, 1]);
    rect1.doesOverlap(rect2).should.eql(0);
    rect1.doesOverlap(rect3).should.eql(0);
    rect1.doesOverlap(rect4).should.eql(0);
    const force1 = overlapRepulsionForce(rect1, rect2, 10);
    force1[0].should.eql(0);
    force1[1].should.be.within(9.99, 10.01);
    const force2 = overlapRepulsionForce(rect1, rect3, 10);
    force2[0].should.eql(0);
    force2[1].should.be.within(4.99, 5.01);
    const force3 = overlapRepulsionForce(rect1, rect4, 10);
    force3[0].should.eql(0);
    force3[1].should.be.within(-0.01, 0.01);
    const force4 = overlapRepulsionForce(rect1, rect5, 10);
    force4[0].should.be.within(-0.01, 0.01);
    force4[1].should.be.within(-0.01, 0.01);
  });

  it('boxes at target radius should have no net force', () => {
    const rect1 = new Rectangle([0, 0], [1, 1]);
    const rect2 = new Rectangle([11, 0], [1, 1]);
    rect1.center.should.eql([0.5, 0.5]);
    rect2.center.should.eql([11.5, 0.5]);
    const force = targetRadiusForce(rect1, rect2, 10);
    force[0].should.be.within(-0.01, 0.01);
    force[1].should.be.within(-0.01, 0.01);
  });

  it('can construct a unit vector', () => {
    Vector.unitVector(0, 2).should.eql([1, 0]);
    Vector.unitVector(1, 2).should.eql([0, 1]);
  });

  it('normalized vector should have length = 1', () => {
    const v = Vector.from([2, 0]);
    const u = v.normalize();
    u.magnitude.should.be.within(1 - EPSILON, 1 + EPSILON);
  });

  it('random unit vector should have length = 1', () => {
    const u = Vector.randomUnitVector(2);
    u.magnitude.should.be.within(1 - EPSILON, 1 + EPSILON);
  });

  it('can compute intersection of line with rectangle', () => {
    const rect = new Rectangle([100, 100], [100, 100]);
    {
      const intersect = rect.lineIntersect([0, 150], [1, 0]);
      intersect[0].should.be.within(99.99, 100.01);
      intersect[1].should.eql(150);
    }
    {
      const intersect = rect.lineIntersect([150, 0], [0, 1]);
      intersect[0].should.eql(150);
      intersect[1].should.be.within(99.99, 100.01);
    }
    {
      const intersect = rect.lineIntersect([0, 0], [1, 1]);
      intersect[0].should.be.within(99.99, 100.01);
      intersect[1].should.be.within(99.99, 100.01);
    }
    {
      const intersect = rect.lineIntersect([0, 150], [-1, 0]);
      (intersect === null).should.be.true;
    }
  });

  it('can add a second box to the graph', async () => {
    await delayOrWait(500);
    const v = graph.addVertex('v1', 'box2');
    v.setProperty('prop', 'value');
  });

  it('can add an edge to the graph', async () => {
    await delayOrWait(500);
    graph.addEdge('e1', 'box1', 'box2', 1);
  });

  it('runs until reaching equilibrium', async () => {
    await graph.runUntilEquilibrium();
  });

  it('can add 10 random nodes', async () => {
    for (let i = 3; i <= 10; i++) {
      await delayOrWait(200);
      const v = graph.addVertex('v2', `box${i}`);
      v.setProperty('prop', 'value');
    }
  });

  it('can add 10 random edges', async () => {
    await delayOrWait(500);
    for (let i = 1; i <= 10; i++) {
      let from;
      let to;
      do {
        from = graph.nodes[Math.floor(Math.random() * graph.nodes.length)];
        to = graph.nodes[Math.floor(Math.random() * graph.nodes.length)];
      } while (from.name === to.name && !graph.getEdge('one', from.name, to.name));
      await delayOrWait(200);
      graph.addEdge('one', from.name, to.name, Math.floor(Math.random() * 100) / 100);
    }
  });

  it.skip('can add 10 more random nodes', async () => {
    for (let i = 11; i <= 20; i++) {
      await delayOrWait(200);
      const v = graph.addVertex('v3', `box${i}`);
      v.setProperty('prop', Math.random() * 10000);
    }
  });

  it('can add 10 more random edges', async () => {
    await delayOrWait(500);
    for (let i = 11; i <= 20; i++) {
      let from;
      let to;
      do {
        from = graph.nodes[Math.floor(Math.random() * graph.nodes.length)];
        to = graph.nodes[Math.floor(Math.random() * graph.nodes.length)];
      } while (from.name === to.name && !graph.getEdge('two', from.name, to.name));
      await delayOrWait(200);
      graph.addEdge('two', from.name, to.name, Math.floor(Math.random() * 100) / 100);
    }
  });
});

mochaRun();
